from __future__ import absolute_import, print_function

import hashlib
import inspect
import logging
import os
import sys
from abc import abstractmethod

import sqlparse
from decorator import decorator
from jinja2 import StrictUndefined, Template

from omniduct.caches.base import cached_method
from omniduct.duct import Duct
from omniduct.utils.debug import logger, logging_scope
from omniduct.utils.docs import quirk_docs
from omniduct.utils.magics import (MagicsProvider, process_line_arguments,
                                   process_line_cell_arguments)

from . import cursor_formatters

logging.getLogger('requests').setLevel(logging.WARNING)


def cache_serializer(format):
    return DatabaseClient.CURSOR_FORMATTERS[format].serialize


def cache_deserializer(format):
    return DatabaseClient.CURSOR_FORMATTERS[format].deserialize


@decorator
def render_statement(method, self, statement, *args, **kwargs):
    if kwargs.pop('template', True):
        statement = self.template_render(
            statement,
            context=kwargs.pop('context', {}),
            by_name=False,
        )
    return method(self, statement, *args, **kwargs)


class DatabaseClient(Duct, MagicsProvider):
    """
    `DatabaseClient` is an abstract subclass of `Duct` that provides a common
    API for all database clients, which in turn will be subclasses of this
    class.

    Note: `DatabaseClient` subclasses are callable, so that one can use
    `DatabaseClient(...)` as a short-hand for `DatabaseClient.query(...)`.

    Class Attributes:
        DUCT_TYPE (`Duct.Type`): The type of `Duct` protocol implemented by this class.
        DEFAULT_PORT (int): The default port for the database service (defined
            by subclasses).
        CURSOR_FORMATTERS (dict<str, CursorFormatter): asdsd
        DEFAULT_CURSOR_FORMATTER (str): ...
    """

    DUCT_TYPE = Duct.Type.DATABASE
    DEFAULT_PORT = None

    CURSOR_FORMATTERS = {
        'pandas': cursor_formatters.PandasCursorFormatter,
        'hive': cursor_formatters.HiveCursorFormatter,
        'csv': cursor_formatters.CsvCursorFormatter,
        'tuple': cursor_formatters.TupleCursorFormatter,
        'dict': cursor_formatters.DictCursorFormatter,
        'raw': cursor_formatters.RawCursorFormatter,
    }
    DEFAULT_CURSOR_FORMATTER = 'pandas'

    @quirk_docs('_init', mro=True)
    def __init__(self, **kwargs):
        """
        templates (dict): A dictionary of name to template mappings. Additional
            templates can be added using `.template_add`.
        template_context (dict): The default template context to use when
            rendering templates.
        """
        Duct.__init_with_kwargs__(self, kwargs, port=self.DEFAULT_PORT)

        self._templates = kwargs.pop('templates', {})
        self._template_context = kwargs.pop('template_context', {})
        self._sqlalchemy_engine = None
        self._sqlalchemy_metadata = None

        self._init(**kwargs)

    @abstractmethod
    def _init(self):
        pass

    def __call__(self, query, **kwargs):
        """
        Allow use of `DatabaseClient(...)` as a short-hand for
        `DatabaseClient.query()`.
        """
        return self.query(query, **kwargs)

    # Querying
    @classmethod
    def statements_split(cls, statements):
        """
        This classmethod converts a single string containing one or more SQL
        statements into an iterator of strings, each corresponding to one SQL
        statement. If the statement's language is not to be SQL, this method
        should be overloaded appropriately.

        Parameters:
            statements (str): A string containing one or more SQL statements.

        Returns:
            iterator<str>: An iterator of SQL statements.
        """
        for statement in sqlparse.split(statements):
            statement = statement.strip()
            if statement.endswith(';'):
                statement = statement[:-1].strip()
            if statement:  # remove empty statements
                yield statement

    @classmethod
    def statement_cleanup(cls, statement):
        """
        This classmethod takes an SQL statement and reformats it by consistently
        removing comments and replacing all whitespace. It is used by the
        `query` method to avoid functionally identical queries hitting different
        cache kets. If the statement's language is not to be SQL, this method
        should be overloaded appropriately.

        Parameters:
            statement (str): The statement to be reformatted/cleaned-up.

        Returns:
            str: The new statement, consistently reformatted.
        """
        statement = sqlparse.format(statement, strip_comments=True, reindent=True)
        statement = os.linesep.join([line for line in statement.splitlines() if line])
        return statement

    @classmethod
    def statement_hash(cls, statement, cleanup=True):
        """
        This classmethod is used to determine the hash used to identify query
        statements to the cache (if configured).

        Parameters:
            statement (str): A string representation of the statement to be
                hashed.
            cleanup (bool): Whether the statement should first be consistently
                reformatted using `statement_cleanup`.

        Returns:
            str: The hash used to identify a statement to the cache.
        """
        if cleanup:
            statement = cls.statement_cleanup(statement)
        if sys.version_info.major == 3 or sys.version_info.major == 2 and isinstance(statement, unicode):  # noqa: F821
            statement = statement.encode('utf8')
        return hashlib.sha256(statement).hexdigest()

    @render_statement
    @quirk_docs('_execute')
    def execute(self, statement, cleanup=True, wait=True, cursor=None, **kwargs):
        """
        This method executes a given statement against the relevant database,
        returning the results as a standard DBAPI2 compatible cursor. Where
        supported by database implementations, this cursor can the be used
        in future executions, by passing it as the `cursor` keyword argument.

        Parameters:
            statement (str): The statement to be executed by the query client
                (possibly templated).
            cleanup (bool): Whether statement should be cleaned up before
                computing the hash used to cache results.
            wait (bool): Whether the cursor should be returned before the
                server-side query computation is complete and the relevant
                results downloaded.
            cursor (DBAPI2 cursor):  Rather than creating a new cursor, execute
                the statement against the provided cursor.
            **kwargs (dict): Extra keyword arguments to be passed on to
                `_execute`, as implemented by subclasses.
            template (bool): Whether the statement should be treated as a Jinja2
                template. [Used by `render_statement` decorator.]
            context (dict): The context in which the template should be
                evaluated (a dictionary of parameters to values). [Used by
                `render_statement` decorator.]

        Returns:
            DBAPI2 cursor: A DBAPI2 compatible cursor instance.
        """

        self.connect()

        statements = self.statements_split(statement)
        statements = [self.statement_cleanup(stmt) if cleanup else stmt for stmt in statements]
        assert len(statements) > 0, "No non-empty statements were provided."

        for statement in statements[:-1]:
            cursor = self.connect()._execute(statement, cursor=cursor, wait=True, **kwargs)
        cursor = self.connect()._execute(statements[-1], cursor=cursor, wait=wait, **kwargs)

        return cursor

    @logging_scope("Query", timed=True)
    @render_statement
    @cached_method(
        id_str=lambda self, kwargs: "{}:\n{}".format(kwargs['format'], self.statement_hash(kwargs['statement'], cleanup=kwargs.get('cleanup', True))),
        format=lambda self, kwargs: kwargs['format'] if kwargs['format'] is not None else self.DEFAULT_CURSOR_FORMATTER,
        serializer=cache_serializer,
        deserializer=cache_deserializer
    )
    def query(self, statement, format=None, format_opts={}, **kwargs):
        """
        This method executes a statement against the database using
        `DatabaseClient.execute()`, and then collects the results before
        returning them formatted as nominated; optionally (and by default)
        caching the result if a cache is configured.

        Parameters:
            statement (str): The statement to be executed by the query client
                (possibly templated).
            format (str): A subclass of CursorFormatter, or one of: 'pandas',
                'hive', 'csv', 'tuple' or 'dict'. Defaults to
                `self.DEFAULT_CURSOR_FORMATTER`.
            format_opts (dict): A dictionary of format-specific options.
            **kwargs (dict): Additional arguments to pass on to
                `DatabaseClient.execute()`.
            use_cache (bool): True (default) or False. Whether to use the cache
                (if present). [Used by `cached_method` decorator.]
            renew (bool): True or False (default). If cache is being used, renew
                it before returning stored value. [Used by `cached_method`
                decorator.]

        Returns:
            The results of the query formatted as nominated.
        """
        cursor = self.execute(statement, wait=True, template=False, **kwargs)

        # Some DBAPI2 cursor implementations error if attempting to extract
        # data from an empty cursor, and if so, we simply return None.
        if self._cursor_empty(cursor):
            return None

        formatter = self._get_formatter(format, cursor, **format_opts)
        return formatter.dump()

    def stream(self, statement, format=None, format_opts={}, batch=None, **kwargs):
        """
        This method executes a statement against the database, and streams
        results from the resulting cursor object as an iterator over objects
        of the nominated format. If `batch` is not `None`, then the iterator
        will be over lists of size `batch`.

        Parameters:
            statement (str): The statement to be executed against the database.
            format (str): A subclass of CursorFormatter, or one of: 'pandas',
                'hive', 'csv', 'tuple' or 'dict'. Defaults to
                `self.DEFAULT_CURSOR_FORMATTER`.
            format_opts (dict): A dictionary of format-specific options.
            batch (int): If not `None`, the number of rows from the resulting
                cursor to be returned at once.
            **kwargs (dict): Additional keyword arguments to pass onto
                `DatabaseClient.execute`.

        Returns:
            iterator: An iterator over objects of the nominated format or, if
                batched, a list of such objects.
        """
        cursor = self.execute(statement, wait=True, **kwargs)
        formatter = self._get_formatter(format, cursor, **format_opts)

        for row in formatter.stream(batch=batch):
            yield row

    def _get_formatter(self, formatter, cursor, **kwargs):
        formatter = formatter or self.DEFAULT_CURSOR_FORMATTER
        if not (inspect.isclass(formatter) and issubclass(formatter, cursor_formatters.CursorFormatter)):
            assert formatter in self.CURSOR_FORMATTERS, "Invalid format '{}'. Choose from: {}".format(formatter, ','.join(self.CURSOR_FORMATTERS.keys()))
            formatter = self.CURSOR_FORMATTERS[formatter]
        return formatter(cursor, **kwargs)

    def stream_to_file(self, statement, file, format='csv', **kwargs):
        """
        This method is a wrapper around `DatabaseClient.stream` that enables the
        iterative writing of cursor results to a file. This is especially useful
        when there are a very large number of results, and loading them all into
        memory would require considerable resources. Note that 'csv' is always
        the default format for this method.

        Parameters:
            statement (str): The statement to be executed against the database.
            file (str, file-like-object): The filename where the data should be
                written, or an open file-like resource.
            format (str): The format to be used ('csv' by default). Format
                options can be passed via `**kwargs`.
            **kwargs: Additional keyword arguments to pass onto
                `DatabaseClient.stream`.
        """
        close_later = False
        if isinstance(file, str):
            file = open(file, 'w')
            close_later = True

        try:
            file.writelines(self.stream(statement, format=format, **kwargs))
        finally:
            if close_later:
                file.close()

    def execute_from_file(self, file, **kwargs):
        """
        This method reads the contents of a file and executes them as a string
        statement against against the database.

        Parameters:
            file (str): The path of the file containing the query statement to
                be executed against the database.
            **kwargs (dict): Extra keyword arguments to pass on to
                `DatabaseClient.execute`.

        Returns:
            DBAPI2 cursor: A DBAPI2 compatible cursor instance.
        """
        with open(file, 'r') as f:
            return self.execute(f.read(), **kwargs)

    def query_from_file(self, file, **kwargs):
        """
        This method reads the contents of a file and executes them as a string
        statement against against the database, returning the results of the
        query formatted as nominated (see `DatabaseClient.query` for more
        details).

        Parameters:
            file (str): The path of the file containing the query statement to
                be executed against the database.
            **kwargs (dict): Extra keyword arguments to pass on to
                `DatabaseClient.query`.

        Returns:
            The results of the query formatted as nominated.
        """
        with open(file, 'r') as f:
            return self.query(f.read(), **kwargs)

    @property
    def template_names(self):
        """
        list: A list of names associated with the templates associated with this
        client.
        """
        return list(self._templates)

    def template_add(self, name, body):
        """
        This method adds a template to the internal dictionary of templates
        stored on this `PrestoClient` instance. Templates are interpreted as
        `jinja2` templates. See `.template_render` for more information.

        Parameters:
            name (str): The name of the template.
            body (str): The (typically) multiline body of the template.

        Returns:
            PrestoClient: A reference to this object.
        """
        self._templates[name] = body
        return self

    def template_render(self, name_or_statement, context=None, by_name=False):
        """
        This method renders a template either by retrieving a template
        associated with a provided template name, or by directly rendering the
        template body as passed.

        In addition to the `jinja2` templating syntax, described in more detail
        in the official `jinja2` documentation, a meta-templating extension is
        also provided. This meta-templating allows you to reference other
        reference other templates. For example, if you had two SQL templates
        named 'template_a' and 'template_b', then you could render them into one
        SQL query using (for example):
        ```
        .template_render('''
        WITH
            a AS (
                {{{template_a}}}
            ),
            b AS (
                {{{template_b}}}
            )
        SELECT *
        FROM a
        JOIN b ON a.x = b.x
        ''')
        ```
        Note that template substitution in this way is iterative, so you can
        chain template embedding, provided that such embedding is not recursive.

        Parameters:
            name_or_statement (str): The name of a template (if `by_name` is True)
                or else a string representation of a `jinja2` template.
            context (dict, None): A dictionary to use as the template context.
                If not specified, an empty dictionary is used.
            by_name (bool): `True` if `name_or_statement` should be interpreted as a
                template name, or `False` (default) if `name_or_statement` should be
                interpreted as a template body.

        Returns:
            str: The rendered template.
        """
        if by_name:
            if name_or_statement not in self._templates:
                raise ValueError("No such template of name: '{}'.".format(name_or_statement))
            statement = self._templates[name_or_statement]
        else:
            statement = name_or_statement

        try:
            from sqlalchemy.sql.base import Executable
            if isinstance(statement, Executable):
                statement = str(statement.compile(compile_kwargs={"literal_binds": True}))
        except ImportError:
            pass

        if context is None or context is False:
            context = {}

        template_context = {}
        template_context.update(self._template_context)  # default context
        template_context.update(context)  # context passed in
        intersection = set(self._template_context.keys()) & set(context.keys())
        if intersection:
            logger.warning(
                "The following default template context keys have been overridden "
                "by the local context: {}."
                .format(intersection)
            )

        # Substitute in any other named statements recursively
        while '{{{' in statement or '{{%' in statement:
            statement = Template(statement,
                                 block_start_string='{{%',
                                 block_end_string='%}}',
                                 variable_start_string='{{{',
                                 variable_end_string='}}}',
                                 comment_start_string='{{#',
                                 comment_end_string='#}}',
                                 undefined=StrictUndefined).render(getattr(self, '_templates', {}))

        return Template(statement, undefined=StrictUndefined).render(template_context)

    def execute_from_template(self, name, context=None, **kwargs):
        """
        This method renders and then executes a named template.

        Parameters:
            name (str): The name of the template to be rendered and executed.
            context (dict): The context in which the template should be rendered.
            **kwargs (dict): Additional parameters to pass to `.execute()`.

        Returns:
            DBAPI2 cursor: A DBAPI2 compatible cursor instance.
        """
        statement = self.template_render(name, context, by_name=True)
        return self.execute(statement, **kwargs)

    def query_from_template(self, name, context=None, **kwargs):
        """
        This method renders and then queries the database using a named template.

        Parameters:
            name (str): The name of the template to be rendered and used to query
                the database.
            context (dict): The context in which the template should be rendered.
            **kwargs (dict): Additional parameters to pass to `.query()`.

        Returns:
           The results of the query formatted as nominated.
        """
        statement = self.template_render(name, context, by_name=True)
        return self.query(statement, **kwargs)

    # Uploading data to data store
    @logging_scope('Push', timed=True)
    @quirk_docs('_push')
    def push(self, df, table, if_exists='fail', **kwargs):
        """
        Todo:
            Review the naming of this method.

        This method uploads a local dataframe to the database, creating,
        overwriting or appending to the nominated table.

        Parameters:
            df (pandas.DataFrame): The dataframe to upload into the database.
            table (str): The name of the table into which the dataframe should
                be uploaded.
            if_exists (str): if nominated table already exists: 'fail' to do
                nothing, 'replace' to drop, recreate and insert data into new
                table, and 'append' to add data from this table into the
                existing table.
            **kwargs (dict): Additional keyword arguments to pass onto
                `QueryClient._push`.
        """
        assert if_exists in {'fail', 'replace', 'append'}
        self.connect()._push(df, table, if_exists=if_exists, **kwargs)

    # Table properties

    @abstractmethod
    def _execute(self, statement, cursor=None, wait=True, **kwargs):
        pass

    def _push(self, df, table, if_exists='fail', **kwargs):
        raise NotImplementedError

    def _cursor_empty(self, cursor):
        return False

    @quirk_docs('_table_list')
    def table_list(self, **kwargs):
        """
        Return a list of table names in the data source as a DataFrame. Additional kwargs are
        passed to `QueryClient._table_list`.
        """
        return self._table_list(**kwargs)

    @abstractmethod
    def _table_list(self, **kwargs):
        pass

    @quirk_docs('_table_exists')
    def table_exists(self, table, **kwargs):
        """
        Return boolean if table exists in schema
        """
        return self._table_exists(table=table, **kwargs)

    @abstractmethod
    def _table_exists(self, table, **kwargs):
        pass

    @quirk_docs('_table_desc')
    def table_desc(self, table, **kwargs):
        """
        Describe a table in the data source. Additional kwargs are
        passed to `QueryClient._table_desc`.

        Returns a pandas dataframe of table fields and descriptors.
        """
        return self._table_desc(table, **kwargs)

    @abstractmethod
    def _table_desc(self, table, **kwargs):
        pass

    @quirk_docs('_table_head')
    def table_head(self, table, n=10, **kwargs):
        """
        Show a sample of the data in `table` of the data source. `n` is the number of records
        to show in this sample. The additional `kwargs` are passed on to `QueryClient._table_head`.

        Returns a pandas DataFrame.
        """
        return self._table_head(table, n=n, **kwargs)

    @abstractmethod
    def _table_head(self, table, n=10, **kwargs):
        pass

    @quirk_docs('_table_props')
    def table_props(self, table, **kwargs):
        """
        Return a dataframe of table properties for `table`.
        """
        return self._table_props(table, **kwargs)

    @abstractmethod
    def _table_props(self, table, **kwargs):
        pass

    def _register_magics(self, base_name):
        """
        The following magic functions will be registered (assuming that
        the base name is chosen to be 'hive'):
        - Cell Magics:
            - `%%hive`: For querying the database.
            - `%%hive.execute`: For executing a statement against the database.
            - `%%hive.stream`: For executing a statement against the database,
                and streaming the results.
            - `%%hive.template`: The defining a new template.
            - `%%hive.render`: Render a provided query statement.
        - Line Magics:
            - `%hive`: For querying the database using a named template.
            - `%hive.execute`: For executing a named template statement against
                the database.
            - `%hive.stream`: For executing a named template against the database,
                and streaming the results.
            - `%hive.render`: Render a provided a named template.
            - `%hive.desc`: Describe the table nominated.
            - `%hive.head`: Return the first rows in a specified table.
            - `%hive.props`: Show the properties specified for a nominated table.

        Documentation for these magics is provided online.
        """
        from IPython import get_ipython
        from IPython.core.magic import register_line_magic, register_cell_magic, register_line_cell_magic

        def statement_executor_magic(executor, statement, variable=None, show='head', transpose=False, template=True, context=None, **kwargs):

            ip = get_ipython()

            if context is None:
                context = ip.user_ns

            # Line magic
            if statement is None:
                return self.query_from_template(variable, context=context, **kwargs)

            # Cell magic
            result = getattr(self, executor)(statement, template=template, context=context, **kwargs)

            if variable is not None:
                ip.user_ns[variable] = result

            if executor != 'query':
                if variable is None:
                    return result
                return
            elif variable is None:
                return result

            format = kwargs.get('format', self.DEFAULT_CURSOR_FORMATTER)
            if show == 'head':
                show = 10
            if isinstance(show, int):
                r = result.head(show) if format == 'pandas' else result[:show]
            elif show == 'all':
                r = result
            elif show == 'none':
                return None
            else:
                raise ValueError("Omniduct does not recognise the argument show='{0}' in cell magic.".format(show))

            if format == 'pandas' and transpose:
                return r.T
            return r

        @register_line_cell_magic(base_name)
        @process_line_cell_arguments
        def query_magic(*args, **kwargs):
            return statement_executor_magic('query', *args, **kwargs)

        @register_line_cell_magic("{}.{}".format(base_name, 'execute'))
        @process_line_cell_arguments
        def execute_magic(*args, **kwargs):
            return statement_executor_magic('execute', *args, **kwargs)

        @register_line_cell_magic("{}.{}".format(base_name, 'stream'))
        @process_line_cell_arguments
        def stream_magic(*args, **kwargs):
            return statement_executor_magic('stream', *args, **kwargs)

        @register_cell_magic("{}.{}".format(base_name, 'template'))
        @process_line_arguments
        def template_add(body, name):
            self.template_add(name, body)

        @register_line_cell_magic("{}.{}".format(base_name, 'render'))
        @process_line_cell_arguments
        def template_render_magic(body=None, name=None, context=None, show=True):

            ip = get_ipython()

            if body is None:
                assert name is not None, "Name must be specified in line-mode."
                rendered = self.template_render(name, context=context or ip.user_ns, by_name=True)
            else:
                rendered = self.template_render(body, context=context or ip.user_ns, by_name=False)
                if name is not None:
                    ip.user_ns[name] = rendered

            if show:
                print(rendered)
            else:
                return rendered

        @register_line_magic("{}.{}".format(base_name, 'desc'))
        @process_line_arguments
        def table_desc(table_name, **kwargs):
            return self.table_desc(table_name, **kwargs)

        @register_line_magic("{}.{}".format(base_name, 'head'))
        @process_line_arguments
        def table_head(table_name, **kwargs):
            return self.table_head(table_name, **kwargs)

        @register_line_magic("{}.{}".format(base_name, 'props'))
        @process_line_arguments
        def table_props(table_name, **kwargs):
            return self.table_props(table_name, **kwargs)



from __future__ import absolute_import

from omniduct.utils.debug import logger

from .base import DatabaseClient


class Neo4jClient(DatabaseClient):
    """
    This Duct connects to a Neo4j graph database server using the `neo4j` python
    library.
    """

    PROTOCOLS = ['neo4j']
    DEFAULT_PORT = 7687
    DEFAULT_CURSOR_FORMATTER = 'raw'

    @classmethod
    def statement_cleanup(cls, statement):
        return statement  # base statement cleanup assumes SQL

    def _init(self):
        self.__driver = None

    # Connection
    def _connect(self):
        from neo4j.v1 import GraphDatabase
        logger.info('Connecting to Neo4J graph database ...')
        auth = (self.username, self.password) if self.username else None
        self.__driver = GraphDatabase.driver("bolt://{}:{}".format(self.host, self.port), auth=auth)  # TODO: Add kerberos support

    def _is_connected(self):
        return hasattr(self, '__driver') and self.__driver is not None

    def _disconnect(self):
        logger.info('Disconnecting from Neo4J graph database ...')
        try:
            self.__driver.close()
        except Exception:
            pass
        self.__driver = None

    # Querying
    def _execute(self, statement, cursor=None, wait=True):
        with self.__driver.session() as session:
            result = session.run(statement)

        # hacky: make the result look like a cursor
        result.close = result.detach
        result.fetchall = result.records
        return result

    def _table_exists(self, table, schema=None):
        raise Exception('tables do not apply to the Neo4J graph database')

    def _table_desc(self, table, **kwargs):
        raise Exception('tables do not apply to the Neo4J graph database')

    def _table_head(self, table, n=10, **kwargs):
        raise Exception('tables do not apply to the Neo4J graph database')

    def _table_list(self, table, schema=None):
        raise Exception('tables do not apply to the Neo4J graph database')

    def _table_props(self, table, **kwargs):
        raise Exception('tables do not apply to the Neo4J graph database')



import inspect
import pickle
import sys
from abc import abstractmethod

import six
from decorator import decorator

from omniduct.duct import Duct
from omniduct.utils.config import config
from omniduct.utils.debug import logger
from omniduct.utils.docs import quirk_docs

config.register('cache_fail_hard',
                description='Raise exception if cache fails to save.',
                default=False)


def cached_method(id_str,
                  cache=lambda self: self.cache,
                  id_duct=lambda self, kwargs: "{}.{}".format(self.__class__.__name__, self.name),
                  use_cache=lambda self, kwargs: kwargs.pop('use_cache', True),
                  renew=lambda self, kwargs: kwargs.pop('renew', False),
                  format=lambda self, kwargs: None,
                  serializer=lambda format: pickle.dump,  # Serializers accept obj, file handle and format.
                  deserializer=lambda format: pickle.load):  # Deserializers accept file handle and format.
    @decorator
    def wrapped(method, self, *args, **kwargs):
        if six.PY3 and not hasattr(sys, 'pypy_version_info'):
            arguments = inspect.signature(method).parameters.keys()
        else:
            arguments = inspect.getargspec(method).args
        kwargs.update(dict(zip(list(arguments)[1:], args)))

        _cache = cache(self)
        _use_cache = use_cache(self, kwargs)
        _renew = renew(self, kwargs)
        _format = format(self, kwargs)

        if _cache is None or not _use_cache:
            return method(self, **kwargs)

        _id_duct = id_duct(self, kwargs)
        _id_str = id_str(self, kwargs)

        if _renew or not _cache.has_key(_id_duct, _id_str):  # noqa: has_key is not of a dictionary here
            value = method(self, **kwargs)
            try:
                _cache.set(
                    id_duct=_id_duct,
                    id_str=_id_str,
                    value=value,
                    serializer=serializer(_format)
                )
            except Exception:  # Remove any lingering (perhaps partial) cache files
                _cache.clear(
                    id_duct=_id_duct,
                    id_str=_id_str
                )
                logger.warning("Failed to save results to cache. If needed, please save them manually.")
                if config.cache_fail_hard:
                    raise

            return value

        logger.caveat('Loaded from cache')

        return _cache.get(
            id_duct=_id_duct,
            id_str=_id_str,
            deserializer=deserializer(_format)
        )
    return wrapped


class Cache(Duct):
    """
    `Cache` is an abstract subclass of `Duct` that provides a common
    API for all cache clients, which in turn will be subclasses of this
    class.

    Note: This will likely be refactored soon to be more powerful, including the
    ability to manage the resource usage of the cache.
    """

    DUCT_TYPE = Duct.Type.CACHE

    @quirk_docs('_init', mro=True)
    def __init__(self, **kwargs):
        """
        This is a shim __init__ function that passes all arguments onto
        `self._init`, which is implemented by subclasses. This allows subclasses
        to instantiate themselves with arbitrary parameters.
        """
        Duct.__init_with_kwargs__(self, kwargs)
        self._init(**kwargs)

    @abstractmethod
    def _init(self):
        pass

    @abstractmethod
    def clear(self, id_duct, id_str):
        pass

    @abstractmethod
    def clear_all(self, id_duct=None):
        pass

    @abstractmethod
    def get(self, id_duct, id_str, deserializer=pickle.load):
        pass

    @abstractmethod
    def has_key(self, id_duct, id_str):
        pass

    @abstractmethod
    def keys(self, id_duct):
        pass

    @abstractmethod
    def set(self, id_duct, id_str, value, serializer=pickle.dump):
        pass



# flake8: noqa

from omniduct.duct import Duct
from omniduct.registry import DuctRegistry
from . import protocols
from ._version import *
from .utils.session import is_directly_imported

if is_directly_imported():
    from .session import *



from __future__ import absolute_import

import ast
import logging
import re
import sys

import pandas.io.sql
import six
from future.utils import raise_with_traceback

from omniduct._version import __version__
from omniduct.utils.debug import logger

from .base import DatabaseClient
from .schemas import SchemasMixin


class PrestoClient(DatabaseClient, SchemasMixin):
    """
    This Duct connects to a Facebook Presto server instance using the `pyhive`
    library.

    In addition to the standard `DatabaseClient` API, `PrestoClient` adds a
    `.schemas` descriptor attribute, which enables a tab completion driven
    exploration of a Presto database's schemas and tables.

    Attributes:
        catalog (str): The default catalog to use in database queries.
        schema (str): The default schema/database to use in database queries.
        connection_options (dict): Additional options to pass on to
            `pyhive.presto.connect(...)`.
    """

    PROTOCOLS = ['presto']
    DEFAULT_PORT = 3506

    def _init(self, catalog='default', schema='default', source=None, **connection_options):
        """
        catalog (str): The default catalog to use in database queries.
        schema (str): The default schema/database to use in database queries.
        source (str): The source of this query (by default "omniduct <version>").
            If manually specified, result will be: "<source> / omniduct <version>".
        connection_options (dict): Additional options to pass on to
            `pyhive.presto.connect(...)`.
        """
        self.catalog = catalog
        self.schema = schema
        self.source = source
        self.connection_options = connection_options
        self.__presto = None
        self.connection_fields += ('catalog', 'schema')

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        omniduct_source = 'omniduct {}'.format(__version__)
        if source is None:
            self._source = omniduct_source
        else:
            self._source = "{} / {}".format(source, omniduct_source)

    # Connection

    def _connect(self):
        from pyhive import presto  # Imported here due to slow import performance in Python 3
        from sqlalchemy import create_engine, MetaData
        logging.getLogger('pyhive').setLevel(1000)  # Silence pyhive logging.
        logger.info('Connecting to Presto coordinator...')
        self.__presto = presto.connect(self.host, port=self.port, username=self.username, password=self.password,
                                       catalog=self.catalog, schema=self.schema,
                                       poll_interval=1, source=self.source, **self.connection_options)
        self._sqlalchemy_engine = create_engine('presto://{}:{}/{}/{}'.format(self.host, self.port, self.catalog, self.schema))
        self._sqlalchemy_metadata = MetaData(self._sqlalchemy_engine)

    def _is_connected(self):
        try:
            return self.__presto is not None
        except:
            return False

    def _disconnect(self):
        logger.info('Disconnecting from Presto coordinator...')
        try:
            self.__presto.close()
        except:
            pass
        self.__presto = None
        self._sqlalchemy_engine = None
        self._sqlalchemy_metadata = None
        self._schemas = None

    # Querying
    def _execute(self, statement, cursor=None, wait=True):
        """
        If something goes wrong, `PrestoClient` will attempt to parse the error
        log and present the user with useful debugging information. If that fails,
        the full traceback will be raised instead.
        """
        from pyhive.exc import DatabaseError  # Imported here due to slow import performance in Python 3
        try:
            cursor = cursor or self.__presto.cursor()
            cursor.execute(statement)
            status = cursor.poll()
            if wait:
                logger.progress(0)
                # status None means command executed successfully
                # See https://github.com/dropbox/PyHive/blob/master/pyhive/presto.py#L234
                while status is not None and status['stats']['state'] != "FINISHED":
                    if status['stats'].get('totalSplits', 0) > 0:
                        pct_complete = round(status['stats']['completedSplits'] / float(status['stats']['totalSplits']), 4)
                        logger.progress(pct_complete * 100)
                    status = cursor.poll()
                logger.progress(100, complete=True)
            return cursor
        except (DatabaseError, pandas.io.sql.DatabaseError) as e:
            # Attempt to parse database error, before ultimately reraising the same
            # exception, maintaining the full stacktrace.
            exception, exception_args, traceback = sys.exc_info()

            try:
                message = e.args[0]
                if isinstance(message, six.string_types):
                    message = ast.literal_eval(re.match("[^{]*({.*})[^}]*$", e.message).group(1))

                linenumber = message['errorLocation']['lineNumber'] - 1
                splt = statement.splitlines()
                splt[linenumber] += '   <--  {errorType} ({errorName}) occurred. {message} '.format(**message)
                context = '\n\n[Error Context]\n{}\n'.format('\n'.join([splt[l] for l in range(max(linenumber - 1, 0),
                                                                                               min(linenumber + 2, len(splt)))]))

                class ErrContext(object):

                    def __repr__(self):
                        return context

                # logged twice so that both notebook and console users see the error context
                exception_args.args = [exception_args, ErrContext()]
                logger.error(context)
            except:
                logger.warn(("Omniduct was unable to parse the database error messages. Refer to the "
                             "traceback below for full error details."))

            if isinstance(exception, type):
                exception = exception(exception_args)

            raise_with_traceback(exception, traceback)

    def _push(self, df, table, if_exists='fail', schema=None, **kwargs):
        """
        Additional parameters:
            schema (str): The schema into which the table should be pushed. If
                not specified, the schema will be set to your username.
        """
        return df.to_sql(name=table, con=self._sqlalchemy_engine, index=False,
                         if_exists=if_exists, schema=schema or self.username, **kwargs)

    def _cursor_empty(self, cursor):
        return False

    def _table_list(self, schema=None, like=None, **kwargs):
        cmd = "SHOW TABLES "
        if schema is not None:
            cmd = cmd + " FROM " + schema
        if like is not None:
            cmd = cmd + " LIKE " + like + "'"
        return self.query(cmd, **kwargs)

    def _table_exists(self, table, schema=None):
        return (self.table_list(renew=True, schema=schema)['Table'] == table).any()

    def _table_desc(self, table, **kwargs):
        return self.query("DESCRIBE {0}".format(table), **kwargs)

    def _table_head(self, table, n=10, **kwargs):
        return self.query("SELECT * FROM {} LIMIT {}".format(table, n), **kwargs)

    def _table_props(self, table, **kwargs):
        raise NotImplementedError



import inspect


def get_importing_modules():
    caller_frame = inspect.currentframe()
    first = True
    while caller_frame is not None:
        try:
            host = inspect.getmodule(caller_frame).__name__
            if not first:
                yield host
            else:
                first = False
        except GeneratorExit:
            return
        except:
            pass

        caller_frame = caller_frame.f_back


def is_directly_imported():
    for module in get_importing_modules():
        return module.startswith('IPython')
    return True



import json
import xml.dom.minidom

import requests
from six.moves import http_client

from pywebhdfs import errors
from pywebhdfs.webhdfs import (PyWebHdfsClient, _is_standby_exception,
                               _move_active_host_to_head)


class OmniductPyWebHdfsClient(PyWebHdfsClient):

    def __init__(self, remote=None, namenodes=None, **kwargs):
        self.remote = remote
        self.namenodes = namenodes or []

        PyWebHdfsClient.__init__(self, **kwargs)

        if self.namenodes and 'path_to_hosts' not in kwargs:
            self.path_to_hosts = [('.*', self.namenodes)]

        # Override base uri
        self.base_uri_pattern = kwargs.get('base_uri_pattern', "http://{host}/webhdfs/v1/").format(
            host="{host}")

    @property
    def host(self):
        host = 'localhost' if self.remote else self._host
        return '{}:{}'.format(host, str(self.port))

    @host.setter
    def host(self, host):
        self._host = host

    @property
    def port(self):
        if self.remote:
            return self.remote.port_forward('{}:{}'.format(self._host, self._port))
        return self._port

    @port.setter
    def port(self, port):
        self._port = port

    @property
    def namenodes(self):
        if self.remote:
            return ['localhost:{}'.format(self.remote.port_forward(nn)) for nn in self._namenodes]
        else:
            return self._namenodes

    @namenodes.setter
    def namenodes(self, namenodes):
        self._namenodes = namenodes

    def _make_uri_local(self, uri):
        if not self.remote:
            return uri
        uri = self.remote.get_local_uri(uri)
        return uri

    def get_home_directory(self):
        response = self._resolve_host(requests.get, True, '/', operation='GETHOMEDIRECTORY')
        if response.ok:
            return json.loads(response.content)['Path']
        return '/'

    def _resolve_host(self, req_func, allow_redirect,
                      path, operation, **kwargs):
        """
        internal function used to resolve federation and HA and
        return response of resolved host.
        """
        import requests
        uri_without_host = self._create_uri(path, operation, **kwargs)
        hosts = self._resolve_federation(path)
        for host in hosts:
            uri = uri_without_host.format(host=host)
            try:
                while True:
                    response = req_func(uri, allow_redirects=False,
                                        timeout=self.timeout,
                                        **self.request_extra_opts)

                    if allow_redirect and response.status_code == http_client.TEMPORARY_REDIRECT:
                        uri = self._make_uri_local(response.headers['location'])
                    else:
                        break

                if not allow_redirect and response.status_code == http_client.TEMPORARY_REDIRECT:
                    response.headers['location'] = self._make_uri_local(response.headers['location'])

                if not _is_standby_exception(response):
                    _move_active_host_to_head(hosts, host)
                    return response
            except requests.exceptions.RequestException:
                pass
        raise errors.ActiveHostNotFound(msg="Could not find active host")


class CdhHdfsConfParser(object):
    """
    This class serves to automatically extract HDFS cluster information from
    Cloudera configuration files.
    """

    def __init__(self, fs, conf_path=None):
        """
        Parameters:
            conf_path (str): The path of the configuration file to be parsed.
            fs (FileSystemClient): The filesystem on which the configuration
                file should be found.
        """
        self.fs = fs
        self.conf_path = conf_path or '/etc/hadoop/conf.cloudera.hdfs2/hdfs-site.xml'

    @property
    def config(self):
        if not hasattr(self, '_config'):
            self._config = self._get_config()
        return self._config

    def _get_config(self):
        with self.fs.open(self.conf_path) as f:
            d = xml.dom.minidom.parseString(f.read())

        properties = d.getElementsByTagName('property')

        return {
            prop.getElementsByTagName('name')[0].childNodes[0].wholeText:
                prop.getElementsByTagName('value')[0].childNodes[0].wholeText
            for prop in properties
        }

    @property
    def clusters(self):
        clusters = []
        for key in self.config:
            if key.startswith('dfs.ha.namenodes.'):
                clusters.append(key[len('dfs.ha.namenodes.'):])
        return clusters

    def namenodes(self, cluster):
        namenodes = self.config['dfs.ha.namenodes.{}'.format(cluster)].split(',')
        return [
            self.config['dfs.namenode.http-address.{}.{}'.format(cluster, namenode)]
            for namenode in namenodes
        ]
